#!/usr/bin/env python

import random, time, select, errno

from gearman.compat import *
from gearman.connection import GearmanConnection
from gearman.task import Task, Taskset

class GearmanBaseClient(object):
    class ServerUnavailable(Exception): pass
    class CommandError(Exception): pass
    class InvalidResponse(Exception): pass

    def __init__(self, job_servers, prefix=None, pre_connect=False):
        """
        job_servers = ['host:post', 'host', ...]
        """
        self.prefix = prefix and "%s\t" % prefix or ""
        self.set_job_servers(job_servers, pre_connect)

    def set_job_servers(self, servers, pre_connect = False):
        # TODO: don't shut down dups and shut down old ones gracefully
        self.connections = []
        for serv in servers:
            connection = GearmanConnection(serv)
            if pre_connect:
                try:
                    connection.connect()
                except connection.ConnectionError:
                    pass # TODO: connection IS marked as dead but perhaps we don't want it at all
            self.connections.append( connection )

class GearmanClient(GearmanBaseClient):
    class TaskFailed(Exception): pass

    def __call__(self, func, arg, uniq=None, **kwargs):
        return self.do_task(Task(func, arg, uniq, **kwargs))

    def do_task(self, task):
        """Return the result of the task or raise a TaskFailed exception on failure."""
        def _on_fail():
            raise self.TaskFailed("Task failed")
        task.on_fail.append(_on_fail)
        ts = Taskset( [task] )
        if not self.do_taskset(ts, timeout=task.timeout):
            raise self.TaskFailed("Task timeout")
        return task.result

    def dispatch_background_task(self, func, arg, uniq=None, high_priority=False):
        """Submit a background task and return its handle."""
        task = Task(func, arg, uniq, background=True, high_priority=True)
        ts = Taskset( [task] )
        self.do_taskset( ts )
        return task.handle

    def get_server_from_hash(self, hsh):
        """Return a live connection for the given hash"""
        # TODO: instead of cycling through, should we shuffle the list if the first connection fails or is dead?
        first_idx = hsh % len(self.connections)
        all_dead = all(conn.is_dead for conn in self.connections)
        for idx in range(first_idx, len(self.connections)) + range(0, first_idx):
            conn = self.connections[idx]

            # if all of the connections are dead we should try reconnecting
            if conn.is_dead and not all_dead:
                continue

            try:
                conn.connect() # Make sure the connection is up (noop if already connected)
            except conn.ConnectionError:
                pass
            else:
                return conn

        raise self.ServerUnavailable("Unable to Locate Server")

    def _submit_task(self, task):
        server = self.get_server_from_hash(hash(task))
        if task.background:
            func = "submit_job_bg"
        elif task.high_priority:
            func = "submit_job_high"
        else:
            func = "submit_job"
        server.send_command(func,
            dict(func=self.prefix + task.func, arg=task.arg, uniq=task.uniq))
        server.waiting_for_handles.insert(0,task)
        return server

    def _command_handler(self, taskset, conn, cmd, args):
        # DEBUG and _D( "RECEIVED COMMAND:", cmd, args )

        handle = ('handle' in args) and ("%s//%s" % (conn.hostspec, args['handle'])) or None

        if cmd != 'job_created' and handle:
            task = taskset.get( taskset.handles.get(handle, None), None)
            if not task:
                return
            if task.is_finished:
                raise self.InvalidResponse("Task %s received %s" % (repr(task), cmd))

        if cmd == 'work_complete':
            task.complete(args['result'])
        elif cmd == 'work_fail':
            if task.retries_done < task.retry_count:
                task.retries_done += 1
                task.retrying()
                task.handle = None
                taskset.connections.add(self._submit_task(task))
            else:
                task.fail()
        elif cmd == 'work_status':
            task.status(int(args['numerator']), int(args['denominator']))
        elif cmd == 'job_created':
            task = conn.waiting_for_handles.pop()
            task.handle = handle
            taskset.handles[handle] = hash( task )
            if task.background:
                task.is_finished = True
        elif cmd == 'error':
            raise self.CommandError(str(args)) # TODO make better
        else:
            raise Exception("Unexpected command: %s" % cmd)

    def do_taskset(self, taskset, timeout=None):
        """Execute a Taskset and return True iff all tasks finished before timeout."""

        # set of connections to which jobs were submitted
        taskset.connections = set(self._submit_task(task) for task in taskset.itervalues())

        taskset.handles = {}

        start_time = time.time()
        end_time = timeout and start_time + timeout or 0
        while not taskset.cancelled and not all(t.is_finished for t in taskset.itervalues()):
            timeleft = timeout and end_time - time.time() or 0.5
            if timeleft <= 0:
                taskset.cancel()
                break

            rx_socks = [c for c in taskset.connections if c.readable()]
            tx_socks = [c for c in taskset.connections if c.writable()]
            try:
                rd, wr, ex = select.select(rx_socks, tx_socks, taskset.connections, timeleft)
            except select.error, e:
                # Ignore interrupted system call, reraise anything else
                if e[0] != errno.EINTR:
                    raise

            for conn in ex:
                pass # TODO

            for conn in rd:
                for cmd in conn.recv():
                    self._command_handler(taskset, conn, *cmd)

            for conn in wr:
                conn.send()

        # TODO: should we fail all tasks that didn't finish or leave that up to the caller?

        return all(t.is_finished for t in taskset.itervalues())



import random, sys, select
from time import time
from gearman.client import GearmanBaseClient

class GearmanJob(object):
    def __init__(self, conn, func, arg, handle):
        self.func = func
        self.arg = arg
        self.handle = handle
        self.conn = conn

    def status(self, numerator, denominator):
        self.conn.send_command_blocking("work_status", dict(handle=self.handle, numerator=numerator, denominator=denominator))

    def complete(self, result):
        self.conn.send_command_blocking("work_complete", dict(handle=self.handle, result=result))

    def fail(self):
        self.conn.send_command_blocking("work_fail", dict(handle=self.handle))

    def __repr__(self):
        return "<GearmanJob func=%s arg=%s handle=%s conn=%s>" % (self.func, self.arg, self.handle, repr(self.conn))

class GearmanWorker(GearmanBaseClient):
    def __init__(self, *args, **kwargs):
        kwargs['pre_connect'] = True

        super(GearmanWorker, self).__init__(*args, **kwargs)
        self.abilities = {}

    def register_function(self, name, func, timeout=None):
        name = self.prefix + name
        self.abilities[name] = (func, timeout)
        self._can_do(self.connections, name, timeout)

    def register_class(self, clas, name=None, decorator=None):
        obj = clas
        if not isinstance(clas, type):
            clas = clas.__class__
        name = name or getattr(obj, 'name', clas.__name__)
        for k in clas.__dict__:
            v = getattr(obj, k)
            if callable(v) and k[0] != '_':
                if decorator:
                    v = decorator(v)
                self.register_function("%s.%s" % (name,k), v)

    def _can_do(self, connections, name, timeout=None):
        cmd_name = (timeout is None) and "can_do" or "can_do_timeout"
        cmd_args = (timeout is None) and dict(func=name) or dict(func=name, timeout=timeout)
        for conn in connections:
            conn.send_command(cmd_name, cmd_args)

    def _set_abilities(self, conn):
        for name, args in self.abilities:
            self._can_do(conn, name, args[1])

    def _alive_connections(self):
        random.shuffle(self.connections)
        for conn in self.connections:
            if not conn.connected and not conn.is_dead:
                try:
                    conn.connect()
                except conn.ConnectionError:
                    continue
                else:
                    self._set_abilities(conn)
            if conn.connected:
                yield conn

    def stop(self):
        self.working = False

    def _work_connection(self, conn, hooks):
        conn.send_command("grab_job")
        cmd = ('noop',)
        while cmd and cmd[0] == 'noop':
            cmd = conn.recv_blocking(timeout=0.5)

        if not cmd or cmd[0] == 'no_job':
            return False

        if cmd[0] != "job_assign":
            # TODO: log this, alert someone, SCREAM FOR HELP, AAAHHHHHH
            # if cmd[0] == "error":
            #     msg = "Error from server: %d: %s" % (cmd[1]['err_code'], cmd[1]['err_text'])
            # else:
            #     msg = "Was expecting job_assigned or no_job, received %s" % cmd[0]
            conn.close()
            return False

        job = GearmanJob(conn, **cmd[1])
        try:
            func = self.abilities[cmd[1]['func']][0]
        except KeyError:
            # TODO: log this - received work for unknown func
            return True

        if hooks:
            hooks.start(job)
        try:
            result = func(job)
        except Exception:
            if hooks:
                hooks.fail(job, sys.exc_info())
            job.fail() # TODO: handle ConnectionError
        else:
            if hooks:
                hooks.complete(job, result)
            job.complete(result) # TODO: handle ConnectionError

        return True

    def work(self, stop_if=None, hooks=None):
        self.working = True
        stop_if = stop_if or (lambda *a,**kw:False)
        last_job_time = time()
        while self.working:
            need_sleep = True

            for conn in self._alive_connections():
                try:
                    worked = self._work_connection(conn, hooks)
                except conn.ConnectionError:
                    pass
                else:
                    if worked:
                        last_job_time = time()
                        need_sleep = False

            is_idle = False
            if need_sleep:
                is_idle = True
                for conn in self.connections:
                    conn.send_command("pre_sleep")
                alive = self._alive_connections()
                try:
                    rd, wr, ex = select.select([c for c in alive if c.readable()], [], alive, 10)
                except select.error, e:
                    # Ignore interrupted system call, reraise anything else
                    if e[0] != 4:
                        raise
                    id_idle = True
                else:
                    is_idle = not bool(rd)

            if stop_if(is_idle, last_job_time):
                self.working = False



"""Compatability with older Python releases"""

try:
    all # Python 2.5
except NameError:
    def all(iter):
        for v in iter:
            if not v:
                return False
        return True



import random

class Task(object):
    hooks = ('on_complete', 'on_fail', 'on_retry', 'on_status', 'on_post')

    def __init__(self, func, arg, uniq=None, background=False, high_priority=False,
                 timeout=None, retry_count=0, **kwargs):
        for hook in self.hooks:
            setattr(self, hook, hook in kwargs and [kwargs[hook]] or [])

        self.func          = func
        self.arg           = arg
        self.uniq          = uniq
        self.background    = background
        self.high_priority = high_priority
        self.timeout       = timeout
        self.retry_count   = retry_count

        self.retries_done = 0
        self.is_finished  = False
        self.handle       = None
        self.result       = None
        self._hash        = hash(self.func + (self.uniq == '-' and self.arg or self.uniq or str(random.randint(0,999999))))

    def __hash__(self):
        return self._hash

    def merge_hooks(self, task2):
        for hook in self.hooks:
            getattr(self, hook).extend(getattr(task2, hook))

    def complete(self, result):
        """Mark the job as completed and call on_complete hooks."""
        self.result = result
        for func in self.on_complete:
            func(result)
        self._finished()

    def fail(self):
        """Mark the job as failed and call on_fail hooks."""
        for func in self.on_fail:
            func()
        self._finished()

    def status(self, numerator, denominator):
        """Call on_status hooks"""
        for func in self.on_status:
            func(numerator, denominator)

    def retrying(self):
        """Call on_retry hooks"""
        for func in self.on_retry:
            func()

    def _finished(self):
        """Mark the job as finished and call on_port hooks."""
        self.is_finished = True
        for func in self.on_post:
            func()
        for hook in self.hooks:
            delattr(self, hook) # TODO: perhaps should just clear the lists?

    def __repr__(self):
        return "<Task func='%s'>" % self.func

class Taskset(dict):
    """
    A Taskset is a group of tasks that are to be run all at once. The
    benefit of using a Taskset is allowing multiple tasks to run
    in parallel.
    """

    def __init__(self, tasks=[]):
        super(Taskset, self).__init__((hash(t), t) for t in tasks)
        self.cancelled = False

    def add(self, task):
        self[hash(task)] = task

    def add_task(self, *args, **kwargs):
        self.add(Task(*args, **kwargs))

    def cancel(self):
        self.cancelled = True

    def __or__(self, taskset2):
        for k,v in taskset2.iteritems():
            if k in self:
                self[k].merge_hooks(v)
            else:
                self[k] = v # TODO: should clone the task rather than just making a reference
        return self



import socket, struct, select, errno
from time import time

DEFAULT_PORT = 4730

COMMANDS = {
     1: ("can_do", ["func"]),
    23: ("can_do_timeout", ["func", "timeout"]),
     2: ("cant_do", ["func"]),
     3: ("reset_abilities", []),
    22: ("set_client_id", ["client_id"]),
     4: ("pre_sleep", []),

     6: ("noop", []),
     7: ("submit_job", ["func", "uniq", "arg"]),
    21: ("submit_job_high", ["func", "uniq", "arg"]),
    18: ("submit_job_bg", ["func", "uniq", "arg"]),

     8: ("job_created", ["handle"]),
     9: ("grab_job", []),
    10: ("no_job", []),
    11: ("job_assign", ["handle", "func", "arg"]),

    12: ("work_status", ["handle", "numerator", "denominator"]),
    13: ("work_complete", ["handle", "result"]),
    14: ("work_fail", ["handle"]),

    15: ("get_status", ["handle"]),
    20: ("status_res", ["handle", "known", "running", "numerator", "denominator"]),

    16: ("echo_req", ["text"]),
    17: ("echo_res", ["text"]),

    19: ("error", ["err_code", "err_text"]),

    24: ("all_yours", []),
}
# Create a mapping of function name -> id, args
R_COMMANDS = dict((m[0], (mid, m[1])) for mid,m in COMMANDS.iteritems())

class GearmanConnection(object):
    class ConnectionError(Exception): pass
    class ProtocolError(Exception): pass

    def __init__(self, host, port=DEFAULT_PORT, sock=None, timeout=None):
        """
        A connection to a Gearman server.
        """
        if not sock:
            self.sock = None
            self.connected = False
        else:
            self.sock = sock
            self.connected = True

        if ':' in host:
            host, port = (host.split(':') + [0])[:2]
            port = int(port) or DEFAULT_PORT
            self.addr = (host, port)
        else:
            port = port or DEFAULT_PORT
            self.addr = (host, port)
        self.hostspec = "%s:%d" % (host, port)
        self.timeout  = timeout

        self.is_dead = False
        self._reset_queues()

    def _reset_queues(self):
        self.in_buffer = ""
        self.out_buffer = ""
        self.waiting_for_handles = []

    def fileno(self):
        return self.sock.fileno()

    def writable(self):
        return self.connected and len(self.out_buffer) != 0

    def readable(self):
        return self.connected

    def connect(self):
        """Connect to the server. Raise ConnectionError if connection fails."""

        if self.connected:
            return

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.settimeout( self.timeout )
        try:
            self.sock.connect( self.addr )
        except (socket.error, socket.timeout), e:
            self.sock = None
            self.is_dead = True
            raise self.ConnectionError(str(e))

        self._reset_queues()
        self.is_dead = False
        self.connected = True
        self.sock.setblocking(0)
        self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, struct.pack("L", 1))

    def is_dead():
        def fget(self):
            if self._is_dead and time() >= self._retry_at:
                self._is_dead = False
                # try:
                #     self.connect()
                # except self.ConnectionError:
                #     pass
            return self._is_dead
        def fset(self, value):
            self._is_dead = value
            self._retry_at = value and time() + 60*2 or None # TODO: should be configurable
        return locals()
    is_dead = property(**is_dead())

    def recv(self, size=4096):
        """
        Returns a list of commands: [(cmd_name, cmd_args), ...]
        Raises ConnectionError if the connection dies.
            or ProtocolError if parsing of a command fails.
        """

        assert self.connected

        data = ''
        try:
            data = self.sock.recv(size)
        except socket.error, e:
            if e.args[0] == errno.EWOULDBLOCK:
                return
            raise
        else:
            if not data:
                self.close()
                self.is_dead = True
                raise self.ConnectionError("connection died")

        self.in_buffer += data

        commands = []
        while len(self.in_buffer) >= 12:
            magic, typ, data_len = struct.unpack("!4sLL", self.in_buffer[0:12])
            if len(self.in_buffer) < 12 + data_len:
                break

            if magic != "\x00RES":
                raise self.ProtocolError("Malformed Magic")

            commands.append(self._parse_command(typ, self.in_buffer[12:12 + data_len]))

            self.in_buffer = buffer(self.in_buffer, 12 + data_len)
        return commands

    def send(self):
        assert self.connected

        if len(self.out_buffer) == 0:
            return 0

        try:
            nsent = self.sock.send(self.out_buffer)
        except socket.error, e:
            self.close()
            self.is_dead = True
            raise self.ConnectionError(str(e))

        self.out_buffer = buffer(self.out_buffer, nsent)

        return len(self.out_buffer)

    def send_command(self, name, kwargs={}):
        # DEBUG and _D("GearmanConnection.send_command", name, kwargs)
        pkt = self._pack_command(name, **kwargs)
        self.out_buffer += pkt
        self.send()

    def _parse_command(self, typ, data):
        """
        Returns: function name, argument dictionary
        """
        msg_spec = COMMANDS.get(typ, None)

        if not msg_spec:
            raise self.ProtocolError("Unknown message received: %d" % typ)

        nargs = len(msg_spec[1])
        data = (nargs > 0) and data.split('\x00', nargs-1) or []
        if len(data) != nargs:
            raise self.ProtocolError("Received wrong number of arguments to %s" % msg_spec[0])

        kwargs = dict(
            ((msg_spec[1][i], data[i]) for i in range(nargs))
        )

        return msg_spec[0], kwargs

    def _pack_command(self, name, **kwargs):
        msg = R_COMMANDS[name]
        data = []
        for k in msg[1]:
            v = kwargs.get(k, "")
            if v is None:
                v = ""
            data.append(str(v))
        data = "\x00".join(data)
        return "%s%s" % (struct.pack("!4sII", "\x00REQ", msg[0], len(data)), data)

    def flush(self, timeout=None): # TODO: handle connection failures
        while self.writable():
            try:
                wr = select.select([], [self], [], timeout)[1] # TODO: exc list
            except select.error, e:
                # Ignore interrupted system call, reraise anything else
                if e[0] != 4:
                    raise
            if self in wr:
                self.send()

    def send_command_blocking(self, cmd_name, cmd_args={}, timeout=None):
        self.send_command(cmd_name, cmd_args)
        self.flush(timeout)

    def recv_blocking(self, timeout=None):
        if not hasattr(self, '_command_queue'):
            self._command_queue = []

        if timeout:
            end_time = time() + timeout
        while len(self._command_queue) == 0:
            time_left = timeout and end_time - time() or 0.5
            if time_left <= 0:
                return []

            try:
                rd, wr, ex = select.select([self], self.writable() and [self] or [], [self], time_left)
            except select.error, e:
                # Ignore interrupted system call, reraise anything else
                if e[0] != 4:
                    raise
                rd = wr = ex = []

            for conn in ex:
                pass # TODO

            for conn in rd:
                for cmd in conn.recv():
                    self._command_queue.insert(0, cmd)

            for conn in wr:
                conn.send()

        return len(self._command_queue) > 0 and self._command_queue.pop() or None

    def close(self):
        self.connected = False
        try:
            self.sock.close()
        except:
            pass
        self.sock = None

    def __repr__(self):
        return "<GearmanConnection %s:%d connected=%s dead=%s>" % (self.addr[0], self.addr[1], self.connected, self.is_dead)



#!/usr/bin/env python

from distutils.core import setup

setup(
    name = 'gearman',
    version = '1.0',
    description = 'Gearman Client',
    author = 'Samuel Stauffer',
    author_email = 'samuel@descolada.com',
    url = 'http://danga.com/gearman/',
    packages = ['gearman'],
)



"""
Geraman client

Overview
========

See http://www.danga.com/gearman/ for more about Gearman.

Usage Summary
=============

Example worker::

    worker = GearmanWorker(["127.0.0.1"])
    worker.register_function("echo", lambda job:job.arg)
    worker.work()

Example client::

    Single task::

        client = GearmanClient(["127.0.0.1"])
        res = client.do_task(Task("echo", "foo"))
        assert res == "foo"

    Multiple parallel tasks::

        client = GearmanClient(["127.0.0.1"])
        ts = Taskset([Task(func="echo", arg="foo")])
        client.do_taskset(ts)
        for task in ts:
            assert ts.result == "foo"

"""

__author__ = "Samuel Stauffer <samuel@descolada.com> and Kristopher Tate <kris@bbridgetech.com>"
__version__ = "1.0"
__license__ = "Python"

from gearman.client import GearmanClient
from gearman.worker import GearmanWorker
from gearman.task import Task, Taskset
